%"onlineBTD: Online Block Term Tensor Decomposition"
clear all;clc;close all;
addpath(genpath('.')); 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% To run this code we need supporting tensorlab code :
% https://www.tensorlab.net and add it to path
% addpath('tensorlab'); 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%  Syntetic data creation
 I=50;J=50;K=100; 
 L={[5 6 7];[5 6 7];[5 6 7]};
 [T,fac]= createBTDproblem(I,J,K,L);
 
%% Add noise in original factor, lower the SNR value, more the noise
% Initilization of initial tensor and noisy factor
 SNR= 5;
 initilization_type = 1; % 1: from noisy factors, 2: random initilzation
 initial_batch_size = K *0.1;
 batchsize = 10; % set batch size based on memory available
 
 [U_all,U_initial] = getnoisyFactorandSubset(fac,L,SNR,initial_batch_size,initilization_type);
 
 [T_initial]=T(:,:,1:initial_batch_size);
 
 % BTD ALS for initil tensor 
options.printitn = 2;    % Show convergence progress every 2 iterations.
options.MaxIter = 10000;   % Set convergence max iterations criteria.
options.tol  = 1e-4;     % Set convergence stoping  criteria
options.fac_type  = 1;
[facs_btd_initial,loss_initial,fval_initial,time_initial,mem_initial]= ...
     btd_als(T_initial,U_initial,L,options);

 disp('******************************************************************');
 fprintf('Initial: Approximation Loss: %f, CPU Time: %f, Memory Used: %f  \n',...
    loss_initial(end),time_initial, mem_initial/1048576 );
 disp('******************************************************************'); 
 
 %% online BTD process
itr = 0; % track time and loss
mem_online = 0;
time_online = 0;
facs_btd = facs_btd_initial;

 for i = initial_batch_size+1:batchsize:K
     itr =itr+1;
     fprintf('Slice number: %d\n',i);
     if(i+batchsize-1 > K)
         batchsize=(K-i+1); % help in case last slice size is less than batchsize/
     end
     % create tensor from new slices.
       T_new_slice = T(:,:,i:i+batchsize-1);
       
     % online update 
       st=tic; 
       [facs_intermediate,MemUsed] = onlineBTD(T_new_slice,facs_btd,L,options.fac_type); 
       time_online = time_online + toc(st);
       if mem_online < MemUsed
           mem_online = MemUsed;
       end
     
     % compute approximation loss
     tmp = T(:,:,1:i+batchsize-1);
     [loss_online(itr) ,fval_online(itr) ] = lossCal(tmp,facs_intermediate);
     
     % update factors to give as initilization
     facs_btd = facs_intermediate;

 end
 disp('******************************************************************');
 fprintf('Online: Approximation Loss: %f, CPU Time: %f, Memory Used: %f  \n',...
 loss_online(end),time_online, mem_online/1048576);
 disp('******************************************************************');

 %% full BTD ALS Decomposition

 [facs_btd_full,loss_full,fval_full,time_full,mem_full]= btd_als(T,U_all,L,options);
 
 disp('******************************************************************');
 fprintf('Full BTD: Approximation Loss: %f, CPU Time: %f, Memory Used: %f  \n',...
    loss_full(end),time_full, mem_full/1048576 );
 disp('******************************************************************');
 
 %% plot and compare loss
 
 subplot(2,1,1);
 plot(loss_online,'b--o','LineWidth',2,'MarkerSize',10);
 title('Online BTD Approximation Loss Plot');
 xlabel('Iteration number');
 ylabel('Approximation Loss');
 subplot(2,1,2);
 plot(loss_full(2:end),'R--o','LineWidth',2,'MarkerSize',10);
 title('Full BTD Approximation Loss Plot');
 xlabel('Iteration number');
 ylabel('Approximation Loss');
 hold off;
 
 
